@@ -63,7 +63,7 @@ use compat::{cvt, AllowStd, ContextWaker};
 use futures_io::{AsyncRead, AsyncWrite};
 use futures_util::{
     sink::{Sink, SinkExt},
-    stream::Stream,
+    stream::{FusedStream, Stream},
 };
 use log::*;
 use std::pin::Pin;
@@ -218,6 +218,7 @@ where
 pub struct WebSocketStream<S> {
     inner: WebSocket<AllowStd<S>>,
     closing: bool,
+    ended: bool,
 }
 
 impl<S> WebSocketStream<S> {
@@ -254,6 +255,7 @@ impl<S> WebSocketStream<S> {
         WebSocketStream {
             inner: ws,
             closing: false,
+            ended: false,
         }
     }
 
@@ -311,6 +313,14 @@ where
     fn poll_next(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Option<Self::Item>> {
         #[cfg(feature = "verbose-logging")]
         trace!("{}:{} Stream.poll_next", file!(), line!());
+
+        // The connection has been closed or a critical error has occurred.
+        // We have already returned the error to the user, the `Stream` is unusable,
+        // so we assume that the stream has been "fused".
+        if self.ended {
+            return Poll::Ready(None);
+        }
+
         match futures_util::ready!(self.with_context(Some((ContextWaker::Read, cx)), |s| {
             #[cfg(feature = "verbose-logging")]
             trace!(
@@ -321,12 +331,27 @@ where
             cvt(s.read_message())
         })) {
             Ok(v) => Poll::Ready(Some(Ok(v))),
-            Err(WsError::AlreadyClosed) | Err(WsError::ConnectionClosed) => Poll::Ready(None),
-            Err(e) => Poll::Ready(Some(Err(e))),
+            Err(e) => {
+                self.ended = true;
+                if matches!(e, WsError::AlreadyClosed | WsError::ConnectionClosed) {
+                    Poll::Ready(None)
+                } else {
+                    Poll::Ready(Some(Err(e)))
+                }
+            }
         }
     }
 }
 
+impl<T> FusedStream for WebSocketStream<T>
+where
+    T: AsyncRead + AsyncWrite + Unpin,
+{
+    fn is_terminated(&self) -> bool {
+        self.ended
+    }
+}
+
 impl<T> Sink<Message> for WebSocketStream<T>
 where
     T: AsyncRead + AsyncWrite + Unpin,
@@ -357,34 +382,24 @@ where
     }
 
     fn poll_close(mut self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Result<(), Self::Error>> {
-        if self.closing {
+        let res = if self.closing {
             // After queueing it, we call `write_pending` to drive the close handshake to completion.
-            match (*self).with_context(Some((ContextWaker::Write, cx)), |s| s.write_pending()) {
-                Ok(()) => Poll::Ready(Ok(())),
-                Err(::tungstenite::Error::ConnectionClosed) => Poll::Ready(Ok(())),
-                Err(::tungstenite::Error::Io(err))
-                    if err.kind() == std::io::ErrorKind::WouldBlock =>
-                {
-                    trace!("WouldBlock");
-                    Poll::Pending
-                }
-                Err(err) => Poll::Ready(Err(err)),
-            }
+            (*self).with_context(Some((ContextWaker::Write, cx)), |s| s.write_pending())
         } else {
-            match (*self).with_context(Some((ContextWaker::Write, cx)), |s| s.close(None)) {
-                Ok(()) => Poll::Ready(Ok(())),
-                Err(::tungstenite::Error::ConnectionClosed) => Poll::Ready(Ok(())),
-                Err(::tungstenite::Error::Io(err))
-                    if err.kind() == std::io::ErrorKind::WouldBlock =>
-                {
-                    trace!("WouldBlock");
-                    self.closing = true;
-                    Poll::Pending
-                }
-                Err(err) => {
-                    debug!("websocket close error: {}", err);
-                    Poll::Ready(Err(err))
-                }
+            (*self).with_context(Some((ContextWaker::Write, cx)), |s| s.close(None))
+        };
+
+        match res {
+            Ok(()) => Poll::Ready(Ok(())),
+            Err(::tungstenite::Error::ConnectionClosed) => Poll::Ready(Ok(())),
+            Err(::tungstenite::Error::Io(err)) if err.kind() == std::io::ErrorKind::WouldBlock => {
+                trace!("WouldBlock");
+                self.closing = true;
+                Poll::Pending
+            }
+            Err(err) => {
+                debug!("websocket close error: {}", err);
+                Poll::Ready(Err(err))
             }
         }
     }
